//*********************************************************
//
// Copyright (c) Microsoft. All rights reserved.
//
//*********************************************************

using System;
using System.Collections.Generic;
using Windows.Devices.Enumeration;
using Windows.Devices.HumanInterfaceDevice;
using Windows.Foundation;
using Windows.Storage.Streams;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Navigation;
using SDKTemplate;

namespace CustomTemperatureSensor
{
    /// <summary>
    /// This scenario demonstrates to how register and utilize input reports
    /// </summary>
    public sealed partial class InputReportEvents : SDKTemplate.Common.LayoutAwarePage
    {
        private TypedEventHandler<HidDevice, HidInputReportReceivedEventArgs> inputReportEventHandler;
        Boolean isRegisteredForInputReportEvents;

        // Device that we registered for events with
        private HidDevice registeredDevice;

        // A pointer back to the main page.  This is needed if you want to call methods in MainPage such
        // as NotifyUser()
        private MainPage rootPage = MainPage.Current;

        public InputReportEvents()
        {
            isRegisteredForInputReportEvents = false;
            
            this.InitializeComponent();
        }

        /// <summary>
        /// Invoked when this page is about to be displayed in a Frame.
        ///
        /// We will enable/disable parts of the UI if the device doesn't support it.
        /// 
        /// The custom sensor requires an output report to tell the device to start sending interrupts, so we will start it
        /// just in case we register for the events. We want this to be transparent because it doesn't demonstrate how 
        /// to register for events.
        /// </summary>
        /// <param name="eventArgs">Event data that describes how this page was reached.  The Parameter
        /// property is typically used to configure the page.</param>
        protected override void OnNavigatedTo(NavigationEventArgs eventArgs)
        {
            // We will disable the scenario that is not supported by the device.
            // If no devices are connected, none of the scenarios will be shown and an error will be displayed
            Dictionary<DeviceType, UIElement> deviceScenarios = new Dictionary<DeviceType, UIElement>();
            deviceScenarios.Add(DeviceType.CustomSensor, TemperatureSensorScenario);

            Utilities.SetUpDeviceScenarios(deviceScenarios, DeviceScenarioContainer);

            UpdateRegisterEventButtons();
        }

        /// <summary>
        /// Reset the device into a known state by undoing everything we did to the device (unregister interrupts and
        /// stop the device from creating interrupts)
        /// </summary>
        /// <param name="eventArgs"></param>
        protected override void OnNavigatedFrom(NavigationEventArgs eventArgs)
        {
            UnregisterFromInputReportEvent();
        }

        private async void OnDeviceClosing(EventHandlerForDevice sender, DeviceInformation deviceInformation)
        {
             await rootPage.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, new DispatchedHandler(() =>
                {
                    UpdateRegisterEventButtons();
                }));
        }

        private void RegisterInterruptEvent_Click(Object sender, Windows.UI.Xaml.RoutedEventArgs eventArgs)
        {
            if (EventHandlerForDevice.Current.IsDeviceConnected)
            {
                // Register for input report events

                RegisterForInputReportEvents();

                UpdateRegisterEventButtons();
            }
            else
            {
                Utilities.NotifyDeviceNotConnected();
            }

        }

        private void UnregisterInterruptEvent_Click(Object sender, Windows.UI.Xaml.RoutedEventArgs eventArgs)
        {
            if (EventHandlerForDevice.Current.IsDeviceConnected)
            {
                UnregisterFromInputReportEvent();
 
                UpdateRegisterEventButtons();
            }
            else
            {
                Utilities.NotifyDeviceNotConnected();
            }
        }


        /// <summary>
        /// This callback updates the slider and textbox with the current temperature from the Netduino/LM35 sensor
        ///
        /// This method is called at the frequency specified by the current report interval. 
        /// 
        /// Since all input reports share the same event in HidDevice, the app needs to get the HidInputReport 
        /// from eventArgs.Report and compare report ids and usages with the desired report.
        /// </summary>
        /// <param name="sender">HidDevice that the event is being raised from</param> 
        /// <param name="eventArgs">Contains the HidInputReport that caused the event to raise</param> 
        private async void OnInputReportEvent(HidDevice sender, HidInputReportReceivedEventArgs eventArgs)
        {
            // Retrieve the sensor data
            HidInputReport inputReport = eventArgs.Report;
            IBuffer buffer = inputReport.Data;
            DataReader dr = DataReader.FromBuffer(buffer);
            byte[] bytes = new byte[inputReport.Data.Length];
            dr.ReadBytes(bytes);

            // Render the sensor data
            await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                CurrentReadingText.TextAlignment = TextAlignment.Center;
                CurrentReadingText.Text = bytes[1].ToString();
                TemperatureSlider.Value = (int)bytes[1];

                rootPage.NotifyUser(bytes[1].ToString() + " degrees Fahrenheit, " + bytes[2].ToString() + " millisecond report-interval", NotifyType.StatusMessage);
            });
        }

        /// <summary>
        /// Register for the event that is triggered when the device sends a report to us.
        /// 
        /// All input reports share the same event in HidDevice, so once we register the event, all HidInputReports (regardless of the report id) will
        /// raise the event. Read the comment above OnInputReportEvent for more information on how to distinguish input reports.
        ///
        /// The function also saves the event token so that we can unregister from the event later on.
        /// </summary>
        private void RegisterForInputReportEvents()
        {
            if (!isRegisteredForInputReportEvents)
            {
                inputReportEventHandler = new TypedEventHandler<HidDevice, HidInputReportReceivedEventArgs>(this.OnInputReportEvent);

                // Remember which device we are registering the device with, in case there is a device disconnect and reconnect. We want to avoid unregistering
                // an event handler that is not registered with the device object.
                // Ideally, one should remove the event token (e.g. assign to null) upon the device removal to avoid using it again.
                registeredDevice = EventHandlerForDevice.Current.Device;

                registeredDevice.InputReportReceived += inputReportEventHandler;

                isRegisteredForInputReportEvents = true;
            }
        }

        /// <summary>
        /// Unregisters from the HidDevice's InputReportReceived event that was registered for in the RegisterForInputReportEvents();
        /// </summary>
        private void UnregisterFromInputReportEvent()
        {
            if (isRegisteredForInputReportEvents)
            {
                // Don't unregister event if the device was removed and reconnected because the endpoint event no longer contains our event handler
                if (registeredDevice == EventHandlerForDevice.Current.Device)
                {
                    registeredDevice.InputReportReceived -= inputReportEventHandler;
                }

                registeredDevice = null;
                isRegisteredForInputReportEvents = false;
            }
        }

 
        private void UpdateRegisterEventButtons()
        {
            ButtonRegisterTemperatureDetection.IsEnabled = !isRegisteredForInputReportEvents;
            ButtonUnregisterTemperatureDetection.IsEnabled = isRegisteredForInputReportEvents;
        }
    }
}
