﻿/* Custom USB HID Communication Device - Echo Device
 * Copyright (c) Secret Labs LLC. All Rights Reserved.
 * 
 * Licensed under the Apache 2.0 open source license
 */

// 
// This firmware is based on the UsbHidEchoNetduinoApp that is found on the Netduino forum
// see:  http://forums.netduino.com/index.php?/topic/1514-driverless-pc-netduino-communication-using-usb/
// For license terms see: http://creativecommons.org/licenses/by-sa/3.0/
//

using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using SecretLabs.NETMF.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;
using System.IO.Ports;


namespace NetduinoSensor
{
    public class Program
    {
        public static void Main()
        {
            OutputPort led = new OutputPort(Pins.ONBOARD_LED, false);
            Sensor sensor = new Sensor();

            const int minRange = 0;  // minimum voltage range
            const int maxRange = 3300; // maximum voltage range in milliVolts

            var voltsPin = new AnalogInput(Pins.GPIO_PIN_A0);    // A0 monitors Vout from the LM35
            voltsPin.SetRange(minRange, maxRange);               // The output range is 0 to 3300 milliVolts

            double milliVolts = 0.0;
            double tempC = 0.0;
            double tempF = 0.0;
            int simpleTemp = 0; // We return an integer temperature in the input report
            int CurrentInterval = 250; // 250 millisecond intervals is our default
            int RequestedInterval = 0; // received from host
            int i = 0; // loop counter
            int[] tempArray = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; //used to record the last twelve readings
            int currentTemp = 0; // most recent reading
            bool firstReading = true;

            // We pause for 5 seconds after invokng Open to ensure that
            // the method has completed.
            if (sensor.Open())
            {
                for (i = 0; i < 5; i++)
                {
                    led.Write(true);
                    Thread.Sleep(1000);
                    led.Write(false);
                }
            }

            while (true)
            {
                // Retrieve the current temperature reading

                milliVolts = (double)voltsPin.Read();  // Read returns a value in the specified range
                tempC = milliVolts / 10.0;             // Sensor returns 10mV per degree Centigrade 
                tempF = 1.8 * tempC + 32;              //Convert to degrees Fahrenheit 
                simpleTemp = (int)tempF;

                // Because there are voltage fluctuations when the external
                // power supply is connected to the Netduino, we use a running
                // average to "smooth" the values.

                if (firstReading)
                {
                    firstReading = false;
                    currentTemp = simpleTemp;
                    for (i = 0; i < 12; i++)
                        tempArray[i] = simpleTemp;

                }
                else 
                {
                    tempArray = Shift(simpleTemp, tempArray);   // Shift the array elements and insert the new temp
                    currentTemp = Average(tempArray);           // Compute a running average of the last 12 readings
                }
 
                RequestedInterval = sensor.Update(currentTemp, CurrentInterval);

                // Check for a possible new interval requested via an
                // output report.

                if (RequestedInterval != 0)
                {
                     CurrentInterval = RequestedInterval;
                }
 
                led.Write(true);
                Thread.Sleep(CurrentInterval);
                led.Write(false);
            }
        }

        // This method shifts the elements of the temperature array left 
        // and then inserts the new temperature-reading into the last (or 12th) slot
        // It returns the updated array.

        static int[] Shift(int temperature, int[] tempArray)
        {
            int i = 0;
 
            // Shift array contents left by one element
            for (i = 0; i < (tempArray.Length - 1); i++)
            {
                tempArray[i] = tempArray[i + 1];
            }
            tempArray[tempArray.Length - 1] = temperature;

            return tempArray;
        }

        // This method computes the average temperature from the
        // most recent values in the array of readings.

        static int Average(int[] tempArray)
        {
            int i = 0;
            int currentTemp = 0;

            // Sum the temperatures in the array
            for (i = 0; i < tempArray.Length; i++)
            {
                currentTemp += tempArray[i];
            }

            // Compute the average reading
            currentTemp = currentTemp / 12;

            // Return the average reading
            return currentTemp;
        }
    }
}
