/* Custom USB HID Communication Device - Echo Device
 * Copyright (c) Secret Labs LLC. All Rights Reserved.
 * 
 * Licensed under the Apache 2.0 open source license
 */

// 
// The starting point for this firmware was the UsbHidEchoNetduinoApp sample that is found on the Netduino forum
// see:  http://forums.netduino.com/index.php?/topic/1514-driverless-pc-netduino-communication-using-usb/
// For license terms see: http://creativecommons.org/licenses/by-sa/3.0/
//
//

using System;
using Microsoft.SPOT;
using System.Threading;
using Microsoft.SPOT.Hardware.UsbClient;
using Microsoft.SPOT.Hardware; 

namespace NetduinoSensor
{
    struct InputReport
    {
        public byte Temperature; // Temperature in degrees Fahrenheit
        public byte Interval;    // report interval (or frequency) in seconds
    }

    struct OutputReport
    {
        public byte Interval; // report interval (or frequency) in seconds
    }

    class Sensor
    {
        protected const int WRITE_ENDPOINT = 1;
        protected const int READ_ENDPOINT = 2; 
        protected const int MAXPOWER_MILLIAMPS = 200;
        protected const int HID_DESCRIPTOR_TYPE = 0x21;

        public UsbStream stream;
        public UsbController usbController;

        protected bool started;

        protected byte[] hidGenericReportDescriptorPayload;

        public Sensor()
        {

            hidGenericReportDescriptorPayload = new byte[]
            {
                0x06,0x55,0xFF,     //HID_USAGE_PAGE_VENDOR_DEFINED
                0x09,0xA5,          //HID_USAGE (vendor_defined)

                0xA1,0x01,          //HID_COLLECTION(Application),

                // Input report (device-transmits)
                0x09,0xA7,          //HID_USAGE (vendor_defined)
                0x15,0x00,          //HID_LOGICAL_MIN_8(0), // Minimum temp is 0 degrees F
                0x25,0x96,          //HID_LOGICAL_MAX_8(150), // Max supported temp is 150 degrees F
                0x75,0x08,          //HID_REPORT_SIZE(8),
                0x95,0x01,          //HID_REPORT_COUNT(1),
                0x81,0x02,          //HID_INPUT(Data_Var_Abs),

                0x09,0xA8,          //HID_USAGE (vendor_defined)
                0x15,0x4B,          //HID_LOGICAL_MIN_8(75),   // minimum 75 milliseconds
                0x25,0xFF,          //HID_LOGICAL_MAX_8(255),  // maximum 255 milliseconds
                0x75,0x08,          //HID_REPORT_SIZE(8), 
                0x95,0x01,          //HID_REPORT_COUNT(1), 
                0x81,0x02,          //HID_INPUT(Data_Var_Abs),

                // Output report (device-receives)
                0x09,0xA9,          //HID_USAGE (vendor_defined)
                0x15,0x4B,          //HID_LOGICAL_MIN_8(75),   // minimum 75 milliseconds
                0x25,0xFF,          //HID_LOGICAL_MAX_8(255),  // maximum 255 milliseconds
                0x75,0x08,          //HID_REPORT_SIZE(8), 
                0x95,0x01,          //HID_REPORT_COUNT(1), 
                0x91,0x02,          //HID_OUTPUT(Data_Var_Abs),

                0xC0                //HID_END_COLLECTION

            };
        }

        public bool Open()
        {
            bool succeed = true;

            started = false;

            UsbController[] usbControllers = UsbController.GetControllers();

            if (usbControllers.Length < 1)
            {
                succeed = false;
            }

            if (succeed)
            {
                usbController = usbControllers[0];

                try
                {
 
                    succeed = ConfigureHID();

                    if (succeed)
                    {
                        succeed = usbController.Start();
                    }

                    if (succeed)
                    {
                        stream = usbController.CreateUsbStream(WRITE_ENDPOINT, READ_ENDPOINT);
                    }
                }
                catch (Exception)
                {
                    succeed = false;
                }
            }

            started = true;
            return succeed;
        }

        public bool Close()
        {
            bool succeed = true;

            if (started)
            {
                succeed = usbController.Stop();

                if (succeed)
                {
                    started = false;
                }
            }

            return succeed;
        }

        protected bool ConfigureHID()
        {
            Configuration configuration = new Configuration();
            Configuration.DeviceDescriptor deviceDescriptor = new Configuration.DeviceDescriptor(0x16C0, 0x0012, 0x0100);
            deviceDescriptor.bcdUSB = 0x110;
            deviceDescriptor.bDeviceClass = 0;  
            deviceDescriptor.bDeviceSubClass = 0; 
            deviceDescriptor.bMaxPacketSize0 = 8;
            deviceDescriptor.iManufacturer = 1;
            deviceDescriptor.iProduct = 2;
            deviceDescriptor.iSerialNumber = 0;

            byte descriptorLength;

            if (hidGenericReportDescriptorPayload.Length == 0 || hidGenericReportDescriptorPayload.Length <= byte.MaxValue)
            {
                descriptorLength = (byte)hidGenericReportDescriptorPayload.Length;
            }
            else
            {
                return false;
            }

            byte[] hidClassDescriptorPayload = new byte[]
            {
                0x01, 0x11,     // bcdHID (v1.01) || 0x01, 0x01, changed to 0x01, 0x11, 
                0x00,           // bCountryCode
                0x01,           // bNumDescriptors
                0x22,           // bDescriptorType (report)
                descriptorLength, 0x00      // wDescriptorLength (report descriptor size in bytes)
            };

            // HID class descriptor
            Configuration.ClassDescriptor hidClassDescriptor = new Configuration.ClassDescriptor(HID_DESCRIPTOR_TYPE, hidClassDescriptorPayload);
            // write endpoint
            Configuration.Endpoint writeEndpoint = new Configuration.Endpoint(WRITE_ENDPOINT, 
                Configuration.Endpoint.ATTRIB_Interrupt | Configuration.Endpoint.ATTRIB_Write);
            writeEndpoint.wMaxPacketSize = 16;   // packet size: 16 bytes
            writeEndpoint.bInterval = 10;         // interval: 10ms

            // read endpoint
            Configuration.Endpoint readEndpoint = new Configuration.Endpoint(READ_ENDPOINT, 
                Configuration.Endpoint.ATTRIB_Interrupt | Configuration.Endpoint.ATTRIB_Read);
            readEndpoint.wMaxPacketSize = 16;   // packet size: 16 bytes
            readEndpoint.bInterval = 10;         // interval: 10ms

            Configuration.UsbInterface usbInterface = new Configuration.UsbInterface(0, new Configuration.Endpoint[] { writeEndpoint, readEndpoint });
            usbInterface.classDescriptors = new Configuration.ClassDescriptor[] { hidClassDescriptor };
            usbInterface.bInterfaceClass = 0x03; // Must be 0x03 for any HID device. hid1_11.pdf pp.7 
            usbInterface.bInterfaceSubClass = 0x00; // Sensors do not support a boot device, must be zero. pp. 8
            usbInterface.bInterfaceProtocol = 0x00; // No interface protocol is supported. hid1_11.pdf pp. 9

            // configuration descriptor
            Configuration.ConfigurationDescriptor configurationDescriptor = new Configuration.ConfigurationDescriptor(MAXPOWER_MILLIAMPS, new Configuration.UsbInterface[] { usbInterface });
            configurationDescriptor.bmAttributes = Configuration.ConfigurationDescriptor.ATTRIB_Base | Configuration.ConfigurationDescriptor.ATTRIB_SelfPowered;

            Configuration.GenericDescriptor hidGenericReportDescriptor = new Configuration.GenericDescriptor(0x81, 0x2200, hidGenericReportDescriptorPayload);
            hidGenericReportDescriptor.bRequest = 0x06; // GET_DESCRIPTOR
            hidGenericReportDescriptor.wIndex = 0x00; // INTERFACE 0 (zero)

            Configuration.StringDescriptor manufacturerNameStringDescriptor = new Configuration.StringDescriptor(1, "Microsoft Corporation");
            Configuration.StringDescriptor productNameStringDescriptor = new Configuration.StringDescriptor(2, "Custom Device");
            Configuration.StringDescriptor displayNameStringDescriptor = new Configuration.StringDescriptor(4, "Netduino Temperature Sensor");
            Configuration.StringDescriptor friendlyNameStringDescriptor = new Configuration.StringDescriptor(5, "Netduino Temp Sensor");

            configuration.descriptors = new Configuration.Descriptor[]
            {
                deviceDescriptor,
                configurationDescriptor, 
                manufacturerNameStringDescriptor, 
                productNameStringDescriptor,
                displayNameStringDescriptor,
                friendlyNameStringDescriptor,
                hidGenericReportDescriptor
            };

            usbController.Configuration = configuration;

            if (usbController.ConfigurationError != UsbController.ConfigError.ConfigOK)
                return false;

            return true;
        }

        public int Update(int iTemperature, int iInterval)
        {
            InputReport inputReport = new InputReport();
            byte Interval = 0;
       
            inputReport.Temperature = (byte)iTemperature;
            inputReport.Interval = (byte)iInterval;

            SendInputReport(inputReport);

            Interval = GetOutputReport();
            return (int)Interval;
        }

        protected byte GetOutputReport()
        {

            byte[] outputReport = new byte[1];
            int bytesRead = 0;

            if (stream.CanRead)
            {
                bytesRead = stream.Read(outputReport, 0, 1);
            }

            if (bytesRead > 0)
                return outputReport[0];
            else
                return 0;
        }

        protected void SendInputReport(InputReport report)
        {
 
            byte[] inputReport = new byte[2];

            inputReport[0] = (byte)report.Temperature;
            inputReport[1] = (byte)report.Interval;

            stream.Write(inputReport, 0, 2);
        }

    }
}
